/*

   rta.cpp

   Copyright (c) Michael Strickland

   GNU General Public License (GPLv3)
   See detailed text in license directory

*/

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstring>
#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <complex>
#include <stdlib.h>

using namespace std;

#include "rta.h"
#include "paramreader.h"

// ------------------------------------------------------------------------------------------------
//  Parameters (default values below are overridden by values in params.txt)
// ------------------------------------------------------------------------------------------------

double fpieb = 1;  // 4 Pi eta / s

double rho0 = 0;  // initial rho; abs(rho0) <= rhoMax

double az02 = 1;  // initial longitudinal anisotropy param az0^2; 1/3 < az0^2 < Infinity; 1 is isotropic

double rhoMax = 10; // max rho; grid covers -rhoMax -> rhoMax

double T0 = 1; // "temperature" at rho0

int num = 1000;  // number of grid points to use internally

int npts = 1000;  // numbers of points to output; npts <= num, reduces output file size

int niters = 20;  // number of iterations

int loadIter = 0; // iteration to load from file; if this is zero, then do not read from file

// ------------------------------------------------------------------------------------------------
//  Computed values
// ------------------------------------------------------------------------------------------------

int i0 = num/2; // default for rho0 = 0
double drho = 2*rhoMax/num; // rho step size
double eb = fpieb/4./M_PI; // etabar = eta/S

double y02 = (3*az02 -1)/2; // default y var
double at02 = az02/y02; // default transverse anisotropy

//
// H2 function; anisotropic energy
//

double H2(double y) {
    if (y>1) return y*(atanh(sqrt(y*y - 1)/y) + y*sqrt(y*y - 1))/sqrt(y*y - 1);
    if (y<1) return y*(atan(sqrt(1 - y*y)/y) + y*sqrt(1 - y*y))/sqrt(1 - y*y);
    return 2;
}

double L0 = T0*sqrt(y02/az02)*pow(H2(sqrt(y02))/2,-0.25);  // default intial Lambda for anisotropic IC

// ------------------------------------------------------------------------------------------------
//  Internal global variables
// ------------------------------------------------------------------------------------------------

dcomp* eList1; // array that will hold the values of T^4
dcomp* eList2; // new values of T^4 generated during iteration will be stored here

// ------------------------------------------------------------------------------------------------

//
// print a line to the screen
//
void print_line() {
    const int dwidth = 152;
    for (int i=0;i<dwidth;i++) cout << "-"; cout << endl;
    return;
}

//
// returns the rho value associated with a grid point i
//
inline double pt(int i) {
    return -rhoMax + i*drho + 0.5*drho;
}

//
// returns the index associated with a given value of rho
//
int index(double rho) {
    return roundf((rho+rhoMax)/drho-0.5);
}

//
// Output scaled energy density to file
//
void outputSnapshot(int idx) {
    fstream out;
    char fname[255];
    sprintf(fname,"output/energy_%d.dat",idx);
    out.open(fname, ios::out);
    for (int i=0; i<num; i+=num/npts)
    out << pt(i) << "\t" << setprecision(12) << real(eList1[i]) << "\t" << imag(eList1[i]) << endl;
    out.close();
    return;
}

//
// Output scaled pi^{eta eta} density to file
//
void outputPiee() {
    fstream out;
    out.open("output/piee.dat", ios::out);
    for (int i=0; i<num; i+=num/npts)
    out << pt(i) << "\t" << setprecision(12) << real(eList2[i]) << "\t" << imag(eList2[i]) << endl;
    out.close();
    return;
}

//
// returns inverse tau_eq
//
inline dcomp teqinv(int i) {
    return pow(eList1[i],0.25)/5./eb;
}

//
// Damping Function
//
dcomp D(double rho, double rhop) {
    // check special cases
    if (rho==rhop) return 1.;
    int i1 = index(rhop);
    int i2 = index(rho);
    if (i1==i2) return 1.;
    // otherwise do the integral
    dcomp temp = 0;
    if (i2>i1) {
      for (int k=i1;k<i2;k++) temp += teqinv(k);
    } else{
      for (int k=i2;k<i1;k++) temp -= teqinv(k);
    }
    temp *= drho;
    return exp(-temp);
}

//
// "Weyl" version of the H function
//
dcomp weylH(double rho, double rhop) {
    if (rho==rhop) return 1.;
    double r = cosh(rhop)/cosh(rho);
    if (r==1) return 1.;
    if (r>1)
      return 0.5*(r*r + r*r*r*acos(1./r)/sqrt(1-1./r/r));
    else 
      return 0.5*(r*r + r*r*r*acosh(1./r)/sqrt(1./r/r-1.));
}

//
// Free-streaming bit
//
dcomp freeStream(double rho) {
    
    double sech = 1/cosh(rho);
    double cosh0 = cosh(rho0);
    double temp = 0;
    double x = 0;
    
    if (az02==1 && abs(rho)==abs(rho0)) {
        //cout << "db 1" << endl;
        return L0*L0*L0*L0;
    }
    
    if (az02==1 && rho0==0) {
        return L0*L0*L0*L0*sech*sech*(1+rho*sech/sinh(rho))/2;
    }
    
    if (cosh(rho) == cosh(rho0)/sqrt(y02)) {
        temp = at02*at02*L0*L0*L0*L0*cosh0*cosh0*sech*sech*(y02 + cosh0*cosh0*sech*sech)/2;
    }
    
    if (cosh(rho) > cosh0/sqrt(y02)) {
        x = sqrt(cosh(rho)*cosh(rho) - cosh0*cosh0/y02);
        temp = atanh(x*sech)*cosh0*cosh0 + y02*x*cosh(rho);
        temp *= sech*sech*sech*cosh0*cosh0*at02*at02*L0*L0*L0*L0/2/x;
    }
    
    if (cosh(rho) < cosh0/sqrt(y02)) {
        x = sqrt(cosh0*cosh0/y02-cosh(rho)*cosh(rho));
        temp = atan(x*sech)*cosh0*cosh0 + y02*x*cosh(rho);
        temp *= sech*sech*sech*cosh0*cosh0*at02*at02*L0*L0*L0*L0/2/x;
    }

    return temp;
}

//
// Integation kernel for the energy density
//
inline dcomp eKernel(double rho, double rhop) {
    return D(rho,rhop)*teqinv(index(rhop))*weylH(rho,rhop);
}

//
// Initial guess for scaled T^4 profile
//
inline dcomp eInit(double rho) {
    return T0*T0*T0*T0*pow(cosh(rho0),8./3.)/pow(cosh(rho),8./3.);
}

//
// Loads eList1 with the initial values from eInit
//
void loadInitE() {
    for (int i=0;i<num;i++) eList1[i] = eInit(pt(i));
}

//
// Loads eList1 from a previous initial condition; expects data in "output/energy_{loadIter}.dat"
//
void readInitE() {
    static int nc = 3; // number of data columns
    char fname[255];
    string* strData;
    sprintf(fname,"output/energy_%d.dat",loadIter);
    ifstream in(fname);
    if (in.is_open()) { 
      cout << "==> Reading initial state from " << fname << endl;
      strData = new string[nc*num];
      for (int i=0;i<nc*num;i++) in >> strData[i];
      for (int i=0;i<num;i++) eList1[i] = dcomp(atof(strData[i*nc+1].c_str()),atof(strData[i*nc+2].c_str()));
    } else {
      cout << "==> Unable to load iteration file " << fname << ".  Exiting." << endl;
      exit(-1);
    }
    delete[] strData;
}

//
// Compute the integral
//
dcomp doInt(int j) {
    dcomp temp = 0;
    // grid spans rho0
    if (pt(j)>rho0) {
      for (int i=i0; i<j; i++)
        temp += eKernel(pt(j),pt(i))*eList1[i];
        temp += eKernel(pt(j),pt(j))*eList1[j]/2.;
    } else {
      for (int i=i0-1; i>j; i--)
        temp -= eKernel(pt(j),pt(i))*eList1[i];
        temp -= eKernel(pt(j),pt(j))*eList1[j]/2.0;
    }
    return temp*drho;
}

//
// One iteration step
//
dcomp doStep(int j) {
    double rho = pt(j);
    return D(rho,rho0)*freeStream(rho) + doInt(j);
}

/* ------------ begin piee routines ----------- */

//
// "Weyl" F function
//
dcomp weylF(double rho, double rhop) {
    if (rho==rhop) return 0;
    double x = cosh(rho)/cosh(rhop);
    if (x==1) return 0;
    if (x>1)
      return (x*sqrt(x*x-1)*(1+2*x*x)+(1-4*x*x)*atanh(sqrt(x*x-1)/x))/(x*x*x*(x*x-1)*sqrt(x*x-1))/6;
    else
      return -(x*sqrt(1-x*x)*(1+2*x*x)+(1-4*x*x)*atan(sqrt(1-x*x)/x))/(x*x*x*(1-x*x)*sqrt(1-x*x))/6;
}

//
// Integation kernel for piee
//
inline dcomp pieeKernel(double rho, double rhop) {
    return D(rho,rhop)*teqinv(index(rhop))*weylF(rho,rhop);
}

//
// Compute the piee integral
//
dcomp doPieeInt(int j) {
    dcomp temp = 0;
    // grid spans rho0
    if (pt(j)>rho0) {
      for (int i=i0; i<j; i++)
        temp += pieeKernel(pt(j),pt(i))*eList1[i];
        temp += pieeKernel(pt(j),pt(j))*eList1[j]/2.;
    } else {
      for (int i=i0-1; i>j; i--)
        temp -= pieeKernel(pt(j),pt(i))*eList1[i];
        temp -= pieeKernel(pt(j),pt(j))*eList1[j]/2.;
    }
    return temp*drho;
}

//
// Free-streaming bit for Pi
//
dcomp freeStreamPI(double rho) {
    
    double sech = 1/cosh(rho);
    double csch = 1/sinh(rho);
    double cosh0 = cosh(rho0);
    double temp = 0;
    double x = 0;
    
    if (az02==1 && abs(rho)==abs(rho0)) {
        //cout << "db 1" << endl;
        return 0;
    }
    
    if (az02==1 && rho0==0) {
        return L0*L0*L0*L0*csch*csch*(2+sech*sech+rho*csch*sech*(-4 + sech*sech))/6;
    }
    
    if (cosh(rho) == cosh(rho0)/sqrt(y02)) {
        temp = -at02*L0*L0*L0*L0*cosh0*cosh0*(at02*cosh0*cosh0 - 2*az02*cosh(rho)*cosh(rho))*sech*sech*sech*sech;
    }
    
    if (cosh(rho) > cosh0/sqrt(y02)) {
        x = sqrt(cosh(rho)*cosh(rho) - cosh0*cosh0/y02);
        temp = at02*L0*L0*L0*L0*cosh0*cosh0*(2*az02*az02*x + at02*cosh0*cosh0*sech*(az02*x*sech + atanh(x*sech)*(-4*az02 + at02*cosh0*cosh0*sech*sech)));
        temp /= 6*x*(-at02*cosh0*cosh0+az02*cosh(rho)*cosh(rho));
    }
    
    if (cosh(rho) < cosh0/sqrt(y02)) {
        x = sqrt(cosh0*cosh0/y02-cosh(rho)*cosh(rho));
        temp = at02*L0*L0*L0*L0*cosh0*cosh0*(2*az02*az02*x + at02*cosh0*cosh0*sech*(az02*x*sech + atan(x*sech)*(-4*az02 + at02*cosh0*cosh0*sech*sech)));
        temp /= 6*x*(-at02*cosh0*cosh0+az02*cosh(rho)*cosh(rho));
    }
    
    return temp;
}


//
// piee RHS
//
dcomp pieeRHS(int j) {
    double rho = pt(j);
    return D(rho,rho0)*freeStreamPI(rho) + doPieeInt(j);
}

//
// compute piee
//
void computePiee() {
    for (int j=0;j<num;j++) eList2[j] = pieeRHS(j); // compute piee from last iteration
}

/* ------------ end piee routines ----------- */

//
// Make n iterations
//
void Iterate(int n) {
    dcomp *tmp; // pointer to double used in pointer swap below
    if (loadIter==0) outputSnapshot(loadIter); // dump initial snapshot to disk
    cout << setw(15) << "--> Iteration " << loadIter << ":";
    cout << setw(15) << setprecision(10) << " 0=" << eList1[0] << "\t num/2=" << eList1[num/2] << "\t num=" << eList1[num-1] << endl;
    for (int i=loadIter;i<n+loadIter;i++) {
        cout << setw(15) << "--> Iteration " << i+1 << ":";
        for (int j=0;j<num;j++) {
            eList2[j] = doStep(j); // update
        }
        cout << setw(15) << setprecision(10) << " 0=" << eList2[0] << "\t num/2=" << eList2[num/2] << "\t num=" << eList2[num-1] << endl;
        // pointer swap
        tmp = eList1;
        eList1 = eList2;
        eList2 = tmp;
        // dump snapshot to disk
        outputSnapshot(i+1);
        // compute any observables necessary
        computePiee(); // loads Piee into eList2 array since it's no longer used
        outputPiee();
    }
}

//
// The main routine
//
int main(int argc, char *argv[])
{
    
    print_line();
    // read parameters from file and command line
    readParametersFromFile("params.txt",1);
    if (argc>1) {
    	print_line();
        cout << "Parameters from commandline" << endl;
        print_line();
        readParametersFromCommandLine(argc,argv,1);
    }	
    // perform any processing of parameters necessary
    processParameters();
    
    /*cout << "L0 = " << L0 << endl;
    for (int i=0;i<21;i++) {
        double tmp = -rhoMax + i*2*rhoMax/20;
        cout << tmp << " : " << freeStreamPI(tmp) << endl;
    }
    exit(0);*/
    
    // allocate memory
    print_line();
    cout << "==> Allocating memory" << endl;
    eList1 = new dcomp[num];
    eList2 = new dcomp[num];

    // load initial condition
    print_line();
    cout << "==> Loading initial guess" << endl;
    if (loadIter==0) loadInitE();
    else readInitE();
    
    // iterate
    print_line();
    cout << "==> Computing" << endl;
    print_line();
    
    Iterate(niters);
    
    // all done, now cleanup
    print_line();
    cout << "==> Freeing memory" << endl;
    delete[] eList1;
    delete[] eList2;

    print_line();

    cout << "==> Done" << endl;
    print_line();
}
